package au.edu.uq.imb.memesuite.servlet;

import au.edu.uq.imb.memesuite.data.Alphabet;
import au.edu.uq.imb.memesuite.db.*;
import au.edu.uq.imb.memesuite.servlet.util.WebUtils;
import au.edu.uq.imb.memesuite.template.HTMLSub;
import au.edu.uq.imb.memesuite.template.HTMLSubGenerator;
import au.edu.uq.imb.memesuite.template.HTMLTemplate;
import au.edu.uq.imb.memesuite.template.HTMLTemplateCache;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import static au.edu.uq.imb.memesuite.servlet.ConfigurationLoader.CACHE_KEY;
import static au.edu.uq.imb.memesuite.servlet.ConfigurationLoader.MOTIF_DB_KEY;


/**
 * Display the available motif databases
 */
public class ShowMotifDBs extends HttpServlet {
  private ServletContext context;
  private HTMLTemplate template;
  private HTMLTemplate categoryTemplate;
  private HTMLTemplate listingTemplate;
  private HTMLTemplate motifTemplate;

  private static Logger logger = Logger.getLogger("au.edu.uq.imb.memesuite.web.motifdb");

  public ShowMotifDBs() { }

  @Override
  public void init() throws ServletException {
    context = this.getServletContext();
    HTMLTemplateCache cache = (HTMLTemplateCache)context.getAttribute(CACHE_KEY);
    template = cache.loadAndCache("/WEB-INF/templates/show_motif_dbs.tmpl");
    categoryTemplate = template.getSubtemplate("content").getSubtemplate("category");
    listingTemplate = categoryTemplate.getSubtemplate("listing");
    motifTemplate = listingTemplate.getSubtemplate("motif");
  }

  @Override
  public void doGet(HttpServletRequest request, HttpServletResponse response)
      throws IOException, ServletException {
    if (request.getParameter("listing") != null) {
      outputXmlListing(response, getId(request, "listing"));
    } else {
      display(response);
    }
  }

  @Override
  public void doPost(HttpServletRequest request, HttpServletResponse response)
      throws IOException, ServletException {
    display(response);
  }

  private void display(HttpServletResponse response)
      throws IOException, ServletException {
    MotifDBList motifDBList = (MotifDBList)context.getAttribute(MOTIF_DB_KEY);
    response.setContentType("text/html");
    HTMLSub out = template.toSub();
    if (motifDBList != null) {
      try {
        out.getSub("content").set("category", new AllCategories(motifDBList));
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error reading motif categories", e);
        out.empty("content");
      }
    } else {
      out.empty("content");
    }
    out.output(response.getWriter());
  }

  private long getId(HttpServletRequest request, String name) throws ServletException {
    String value = request.getParameter(name);
    if (value == null) {
      throw new ServletException("Parameter '" + name + "' was not set.");
    }
    long id;
    try {
      id = Long.parseLong(value, 10);
    } catch (NumberFormatException e) {
      throw new ServletException("Parameter '" + name + "' is not a integer value.", e);
    }
    return id;
  }


  private void outputXmlListing(HttpServletResponse response, long listingId)
      throws IOException {
    MotifDBList motifDBList = (MotifDBList)context.getAttribute(MOTIF_DB_KEY);
    response.setContentType("application/xml");
    PrintWriter out = null;
    try {
      MotifDB motifDB = motifDBList.getMotifListing(listingId);
      out = response.getWriter();
      out.print("<motif_db id=\"" + listingId + "\" alphabet=\"" +
          motifDB.getAlphabet().name() + "\">");
      // name contains HTML, hence can't be escaped or it becomes useless
      out.print("<name><![CDATA[");
      out.print(motifDB.getName());
      out.println("]]></name>");
      // description contains HTML, hence can't be escaped or it becomes useless
      out.print("<description><![CDATA[");
      out.print(motifDB.getDescription());
      out.println("]]></description>");
      for (MotifDBFile file : motifDB.getMotifFiles()) {
        out.println("<file src=\"" +WebUtils.escapeHTML(file.getFileName()) +
            "\" count=\"" + file.getMotifCount() + "\" cols=\"" +
            file.getTotalCols() + "\" min=\"" + file.getMinCols() +
            "\" max=\"" + file.getMaxCols() + "\" />");
      }
      out.println("</motif_db>");
    } catch (SQLException e) {
      throw new IOException(e);
    } finally {
      if (out != null) out.close();
    }

  }

  private class AllCategories extends HTMLSubGenerator<Category> {
    private MotifDBList motifDBList;

    private AllCategories(MotifDBList motifDBList) throws SQLException{
      super(motifDBList.getCategories(false, EnumSet.allOf(Alphabet.class)));
      this.motifDBList = motifDBList;
    }

    @Override
    protected HTMLSub transform(Category item) {
      HTMLSub out = categoryTemplate.toSub();
      try {
        out.set("listing", new AllListingsOfCategory(motifDBList, item.getID()));
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error reading motif listings", e);
        out.empty("listing");
      }
      return out;
    }
  }

  private class AllListingsOfCategory extends HTMLSubGenerator<Listing> {
    private MotifDBList motifDBList;

    public AllListingsOfCategory(MotifDBList motifDBList, long id) throws SQLException {
      super(motifDBList.getListings(id));
      this.motifDBList = motifDBList;
    }

    @Override
    protected HTMLSub transform(Listing item) {
      HTMLSub out = listingTemplate.toSub();
      out.set("name", item.getName());
      out.set("description", item.getDescription());
      try {
        MotifDB motifDB = motifDBList.getMotifListing(item.getID());
        List<HTMLSub> motifSubs = new ArrayList<HTMLSub>();
        for (MotifDBFile motifDBFile : motifDB.getMotifFiles()) {
          HTMLSub motifSub = motifTemplate.toSub();
          motifSub.set("file", motifDBFile.getFileName());
          motifSub.set("count", motifDBFile.getMotifCount());
          motifSub.set("min", motifDBFile.getMinCols());
          motifSub.set("max", motifDBFile.getMaxCols());
          motifSub.set("average", String.format("%.1f", motifDBFile.getAverageCols()));
          motifSubs.add(motifSub);
        }
        out.set("motif", motifSubs);
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error reading motifs of listing", e);
        out.empty("motif");
      }
      return out;
    }
  }

}
