package au.edu.uq.imb.memesuite.servlet;

import au.edu.uq.imb.memesuite.data.Alphabet;
import au.edu.uq.imb.memesuite.db.*;
import au.edu.uq.imb.memesuite.template.HTMLSub;
import au.edu.uq.imb.memesuite.template.HTMLSubGenerator;
import au.edu.uq.imb.memesuite.template.HTMLTemplate;
import au.edu.uq.imb.memesuite.template.HTMLTemplateCache;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import static au.edu.uq.imb.memesuite.servlet.ConfigurationLoader.CACHE_KEY;
import static au.edu.uq.imb.memesuite.servlet.ConfigurationLoader.SEQUENCE_DB_KEY;

/**
 * Display the available sequence databases.
 */
public class ShowSequenceDBs extends HttpServlet {
  private ServletContext context;
  private HTMLTemplate template;
  private HTMLTemplate categoryLiTemplate;
  private HTMLTemplate categoryTemplate;
  private HTMLTemplate listingTemplate;
  private HTMLTemplate versionTemplate;

  private static Logger logger = Logger.getLogger("au.edu.uq.imb.memesuite.web.sequencedb");

  public ShowSequenceDBs() { }

  @Override
  public void init() throws ServletException {
    this.context = this.getServletContext();
    HTMLTemplateCache cache = (HTMLTemplateCache)context.getAttribute(CACHE_KEY);
    template = cache.loadAndCache("/WEB-INF/templates/show_sequence_dbs.tmpl");
    categoryLiTemplate = template.getSubtemplate("content").getSubtemplate("category_li");
    categoryTemplate = template.getSubtemplate("content").getSubtemplate("category");
    listingTemplate = categoryTemplate.getSubtemplate("listing");
    versionTemplate = listingTemplate.getSubtemplate("version");
  }

  @Override
  public void doGet(HttpServletRequest request, HttpServletResponse response)
      throws IOException, ServletException {
    if (request.getParameter("category") != null) {
      outputXmlListingsOfCategory(response, getId(request, "category"),
          isShortOnly(request), allowedAlphabets(request));
    } else if (request.getParameter("listing") != null) {
      outputXmlVersionsOfListing(response, getId(request, "listing"),
          isShortOnly(request), allowedAlphabets(request));
    } else {
      display(response);
    }
  }

  @Override
  public void doPost(HttpServletRequest request, HttpServletResponse response)
      throws IOException, ServletException {
    display(response);
  }

  private void display(HttpServletResponse response)
      throws IOException, ServletException {
    SequenceDBList sequenceDBList = (SequenceDBList)context.getAttribute(SEQUENCE_DB_KEY);
    response.setContentType("text/html");
    HTMLSub out = template.toSub();
    if (sequenceDBList != null) {
      try {
        out.getSub("content").set("category_li", new AllCategoryList(sequenceDBList));
        out.getSub("content").set("category", new AllCategories(sequenceDBList));
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error loading sequence categories", e);
        out.empty("content");
      }
    } else {
      out.empty("content");
    }
    out.output(response.getWriter());
  }

  private long getId(HttpServletRequest request, String name) throws ServletException {
    String value = request.getParameter(name);
    if (value == null) {
      throw new ServletException("Parameter '" + name + "' was not set.");
    }
    long id;
    try {
      id = Long.parseLong(value, 10);
    } catch (NumberFormatException e) {
      throw new ServletException("Parameter '" + name + "' is not a integer value.", e);
    }
    return id;
  }

  private boolean isShortOnly(HttpServletRequest request) throws ServletException{
    String value = request.getParameter("short");
    if (value == null) return false;
    boolean shortOnly;
    try {
      shortOnly = Integer.parseInt(value, 2) != 0;
    } catch (NumberFormatException e) {
      throw new ServletException("Parameter 'short' is not a binary value.", e);
    }
    return shortOnly;
  }

  private EnumSet<Alphabet> allowedAlphabets(HttpServletRequest request) throws ServletException {
    String value = request.getParameter("alphabets");
    if (value == null) return EnumSet.allOf(Alphabet.class);
    EnumSet<Alphabet> allowedAlphabets;
    try {
      allowedAlphabets = SQL.intToEnums(Alphabet.class, Integer.parseInt(value, 10));
    } catch (NumberFormatException e) {
      throw new ServletException("Parameter 'alphabets' is not a bitset.", e);
    } catch (IllegalArgumentException e) {
      throw new ServletException("Parameter 'alphabets' is not a bitset", e);
    }
    return allowedAlphabets;
  }

  private void outputXmlListingsOfCategory(HttpServletResponse response,
      long categoryId, boolean shortOnly, EnumSet<Alphabet> allowedAlphabets)
      throws IOException, ServletException {
    SequenceDBList sequenceDBList = (SequenceDBList)context.getAttribute(SEQUENCE_DB_KEY);
    response.setContentType("application/xml");
    // turn off caching
//    response.setHeader("Cache-Control", "no-cache, no-store, must-revalidate"); // HTTP 1.1.
//    response.setHeader("Pragma", "no-cache"); // HTTP 1.0.
//    response.setDateHeader("Expires", 0); // Proxies.
    PrintWriter out = null;
    try {
      Iterator<Listing> listingView = sequenceDBList.getListings(categoryId, shortOnly, allowedAlphabets).iterator();
      out = response.getWriter();
      out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
      out.println("<listings category=\"" + categoryId + "\" short=\"" +
          (shortOnly ? "1" : "0") +"\" alphabets=\"" +
          SQL.enumsToInt(allowedAlphabets) + "\">");
      while (listingView.hasNext()) {
        Listing listing = listingView.next();
        out.println("<l i=\"" + listing.getID() + "\" n=\"" + listing.getName() +
            "\" a=\"" + SQL.enumsToInt(listing.getAlphabets()) + "\"/>");
      }
      out.println("</listings>");
    } catch (SQLException e) {
      throw new ServletException(e);
    } finally {
      if (out != null) out.close();
    }
  }

  private void outputXmlVersionsOfListing(HttpServletResponse response,
      long listingId, boolean shortOnly, EnumSet<Alphabet> allowedAlphabets)
      throws IOException, ServletException {
    SequenceDBList sequenceDBList = (SequenceDBList)context.getAttribute(SEQUENCE_DB_KEY);
    response.setContentType("application/xml");
    PrintWriter out = null;
    try {
      Iterator<SequenceVersion> versionView = sequenceDBList.getVersions(listingId, shortOnly, allowedAlphabets).iterator();
      out = response.getWriter();
      out.println("<versions listing=\"" + listingId + "\" short=\"" +
          (shortOnly ? "1" : "0") + "\" alphabets=\"" +
          SQL.enumsToInt(allowedAlphabets) + "\">");
      while (versionView.hasNext()) {
        SequenceVersion version = versionView.next();
        out.println("<v i=\"" + version.getEdition() + "\" n=\"" +
            version.getVersion() + "\" a=\"" +
            SQL.enumsToInt(version.getAlphabets()) + "\"/>");
      }
      out.println("</versions>");
    } catch (SQLException e) {
      throw new ServletException(e);
    } finally {
      if (out != null) out.close();
    }

  }

  private class AllCategoryList extends HTMLSubGenerator<Category> {
    private AllCategoryList(SequenceDBList sequenceDBList) throws SQLException{
      super(sequenceDBList.getCategories(false, EnumSet.allOf(Alphabet.class)));
    }

    @Override
    protected HTMLSub transform(Category item) {
      HTMLSub out = categoryLiTemplate.toSub();
      out.set("id", item.getID());
      out.set("name", item.getName());
      return out;
    }
  }

  private class AllCategories extends HTMLSubGenerator<Category> {
    private SequenceDBList sequenceDBList;

    private AllCategories(SequenceDBList sequenceDBList) throws SQLException{
      super(sequenceDBList.getCategories(false, EnumSet.allOf(Alphabet.class)));
      this.sequenceDBList = sequenceDBList;
    }

    @Override
    protected HTMLSub transform(Category item) {
      HTMLSub out = categoryTemplate.toSub();
      out.set("id", item.getID());
      out.set("name", item.getName());
      try {
        out.set("listing", new AllListingsOfCategory(sequenceDBList, item.getID()));
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error loading sequence listings", e);
        out.empty("listing");
      }
      return out;
    }
  }

  private class AllListingsOfCategory extends HTMLSubGenerator<Listing> {
    private SequenceDBList sequenceDBList;

    public AllListingsOfCategory(SequenceDBList sequenceDBList, long id) throws SQLException {
      super(sequenceDBList.getListings(id));
      this.sequenceDBList = sequenceDBList;
    }

    @Override
    protected HTMLSub transform(Listing listing) {
      HTMLSub out = listingTemplate.toSub();
      out.set("name", listing.getName());
      out.set("description", listing.getDescription());
      try {
        out.set("version", new AllVersionsOfListing(sequenceDBList, listing.getID()));
      } catch (SQLException e) {
        logger.log(Level.SEVERE, "Error loading sequence versions", e);
        out.empty("version");
      }
      EnumSet<Alphabet> alphabets = listing.getAlphabets();
      if (!alphabets.contains(Alphabet.RNA)) out.empty("rna");
      if (!alphabets.contains(Alphabet.DNA)) out.empty("dna");
      if (!alphabets.contains(Alphabet.PROTEIN)) out.empty("protein");
      return out;
    }
  }

  private class AllVersionsOfListing extends HTMLSubGenerator<SequenceVersion> {
    private AllVersionsOfListing(SequenceDBList sequenceDBList, long listingId) throws SQLException {
      super(sequenceDBList.getVersions(listingId));
    }

    @Override
    protected HTMLSub transform(SequenceVersion listing) {
      SequenceDB sequenceDB;
      HTMLSub out = versionTemplate.toSub();
      out.set("name", listing.getVersion());
      if ((sequenceDB = listing.getSequenceFile(Alphabet.DNA)) != null) {
        out.getSub("dna").set("description", sequenceDB.getDescription());
      } else {
        out.empty("dna");
      }
      if ((sequenceDB = listing.getSequenceFile(Alphabet.RNA)) != null) {
        out.getSub("rna").set("description", sequenceDB.getDescription());
      } else {
        out.empty("rna");
      }
      if ((sequenceDB = listing.getSequenceFile(Alphabet.PROTEIN)) != null) {
        out.getSub("protein").set("description", sequenceDB.getDescription());
      } else {
        out.empty("protein");
      }
      return out;
    }
  }

}
